#
# Christmas Logic 
# Copyright (C) Jack Whitham 2009
#

from dfl import Distance, Distance_From_Line, Find_P4
from lcm import Least_Common_Multiple
from logic import *
import wx, traceback
import font


class WX_Primitives:
    def __init__(self):
        self.PEN_REGULAR_LINE = wx.Pen((0, 0, 0), 2, wx.SOLID)
        self.PEN_REGULAR_TERMINAL = wx.Pen((0, 0, 0), 1, wx.SOLID)
        self.PEN_REGULAR = wx.Pen((0, 0, 0), 1, wx.SOLID)
        self.PEN_HIGHLIGHT_LINE = wx.Pen(HIGHLIGHT_COLOUR, 
                            2 + HIGHLIGHT_INC, wx.SOLID)
        self.PEN_HIGHLIGHT_TERMINAL = wx.Pen(HIGHLIGHT_COLOUR, 2, wx.SOLID)
        self.PEN_HIGHLIGHT_COMPONENT = wx.Pen(HIGHLIGHT_COLOUR, 2, wx.SOLID)
        self.PEN_TRANSPARENT = wx.Pen((0, 0, 0), 2, wx.TRANSPARENT)
        self.BR_HIGHLIGHT_TERMINAL = wx.Brush(HIGHLIGHT_COLOUR)
        self.BR_HIGHLIGHT_COMPONENT = wx.Brush(HIGHLIGHT_COLOUR)
        self.BR_WHITE = wx.Brush((255, 255, 255))
        self.BR_HOLLOW = wx.Brush((255, 255, 255), wx.TRANSPARENT)
        self.BR_ONE = wx.Brush(COLOUR_ONE)
        self.BR_UNDEFINED = wx.Brush(COLOUR_UNDEFINED)
        self.BR_ZERO = wx.Brush(COLOUR_ZERO)
        self.FONT_CL = font.Make_Font(wx.FONTFAMILY_TELETYPE, 40)
        self.FONT_SMALL = font.Make_Font(wx.FONTFAMILY_TELETYPE, 15)

        self.brush_cache = dict()

    def Get_Brush(self, c):
        if not self.brush_cache.has_key(c):
            self.brush_cache[ c ] = wx.Brush(c)

        return self.brush_cache[ c ]

class Wire_Popup_Menu(wx.Menu):
    def __init__(self, parent, network, click_pos):
        wx.Menu.__init__(self)

        self.parent = parent
        self.network = network
        self.click_pos = click_pos

        self.split = wx.MenuItem(self, wx.NewId(), "Split Here") 
        self.AppendItem(self.split)
        self.Bind(wx.EVT_MENU, self.On_Split, id = self.split.GetId())

        self.delete = wx.MenuItem(self, wx.NewId(), "Delete") 
        self.AppendItem(self.delete)
        self.Bind(wx.EVT_MENU, self.On_Delete, id = self.delete.GetId())

        self.cancel = wx.MenuItem(self, wx.NewId(), "Cancel") 
        self.AppendItem(self.cancel)
        self.Bind(wx.EVT_MENU, self.On_Cancel, id = self.cancel.GetId())

    def On_Split(self, event):
        old_wire = self.network.selected_item
        if not isinstance(old_wire, Wire):
            # What? How did the selected_item change?
            self.On_Cancel(event)
            return

        new_component = Split_Component(self.network, self.click_pos)
        new_terminal = new_component.Get_Terminal()
        new_wire_1 = Wire(self.network, new_terminal, old_wire.t2)
        new_wire_2 = Wire(self.network, new_terminal, old_wire.t1)
        old_wire.Delete()
        self.On_Cancel(event)

    def On_Delete(self, event):
        old_wire = self.network.selected_item
        if isinstance(old_wire, Wire):
            old_wire.Delete()

        self.On_Cancel(event)

    def On_Cancel(self, event):
        self.network.Popup_Done()

class Network:
    def __init__(self):
        self.components = []
        self.terminals = []
        self.all_elements = set()
        self.cell_contents = dict()
        self.selected_item = self.highlight = None
        self.mouse_down = False
        self.creating_wire = None
        self.creating_wire_dest = None
        self.new_elements = set()
        self.map_contents_change = True
        self.network_contents_change = True
        self.popup_menu = None
        self.window_extent = dict()
        self.scroll_point_at = None

        self.clocks = 0
        self.stimulator_sequence = None
        self.stimulator_list = None

        self.string_data = []
        self.string_data_for_name = dict()
        self.goal_major = None
        self.win_flag = False
        self.label_component = None

        self.draw_before_components = []
        self.draw_after_components = []
        self.WX_Disconnect()

    def WX_Disconnect(self):
        self.wxp = None
        self.wxwin = None
        self.win_flag = False

    def WX_Connect(self, wxwin):
        if self.wxp == None:
            self.wxp = WX_Primitives()

        self.wxwin = wxwin
        self.wxwin.Change_Scrollbar(**self.window_extent)

    def Tick(self):
        # Add a clock cycle
        rc = self.Update_Map(0)
        assert rc == None

        if self.stimulator_sequence == None:
            self.__Make_Stimulator_Sequence()

        #print 'tick'
        self.clocks += 1
        for ne in self.components:
            assert ne in self.all_elements
            if isinstance(ne, Clocked_Component):
                if isinstance(ne, Stimulator):
                    ne.pointer = self.clocks
                ne.Tick()
                #print ne, ne.Get_State()

        self.__Simulate()

        # Compute new network state - did anything change?

        if len(self.string_data) == 0:
            self.goal_major = None
            self.win_flag = False
        else:
            self.__Check_Goal()

    def __Check_Goal(self):
        # Winners compare their achievements to their goals - Armonk, NY?
        s = self.string_data[ 0 ]
        gsl = len(s.goal_sequence)

        if self.goal_major != None:
            # nth stage - does this part match?
            assert 0 <= self.goal_major < gsl
            self.goal_major = i = (self.goal_major + 1) % gsl
            for s in self.string_data:
                if s.goal_sequence[ i ] != s.state:
                    self.goal_major = None
                    self.win_flag = False
                    break

            if ((self.goal_major != None) and (i == 0)):
                # Cycle completed. Win!
                self.win_flag = True
                #print "YOU'RE A WINNER!"

        if self.goal_major == None:
            # First stage - does even the first part match?
            self.goal_major = 0
            for s in self.string_data:
                if s.goal_sequence[ 0 ] != s.state:
                    # No, reset
                    self.goal_major = None
                    self.win_flag = False
                    break

    def Get_Guide_Component(self):
        if isinstance(self.label_component, Guide_Component):
            return self.label_component 
        return None

    def Get_Goal_Major(self):
        return self.goal_major

    def Get_Win_Flag(self):
        return self.win_flag

    def Locate(self, (x, y), terminal_only = False):
        half = (1 << CELL_SHIFT) >> 1
        (cx, cy) = Cell((x + half, y + half))
        cells = [ (cx - 1, cy - 1), (cx - 1, cy) , (cx, cy - 1), (cx, cy) ]
        empty = set()
        all_contents = set()
        finder = []

        for (cx, cy) in cells:
            all_contents.update(self.cell_contents.get((cx, cy), empty))

        if self.creating_wire_dest != None:
            all_contents.discard(self.creating_wire)
            all_contents.discard(self.creating_wire_dest)
            all_contents.discard(self.creating_wire_dest.Get_Terminal())

        for ne in all_contents:
            if isinstance(ne, Terminal):
                dist = Distance((x, y), ne.Get_Location())
                if (dist < TERMINAL_RADIUS):
                    finder.append((0, ne))
                else:
                    finder.append((dist * 2.0, ne))

            elif isinstance(ne, Split_Component) and not terminal_only:
                dist = Distance((x, y), ne.Get_Location())
                if dist < SPECIAL_SIZE:
                    finder.append((1, ne))
                else:
                    finder.append((dist, ne))

            elif (isinstance(ne, Base_Component) and (not terminal_only)
            and (not isinstance(ne, Splash_Component))):
                dist = Distance((x, y), ne.Get_Location())
                finder.append((dist, ne))

            elif isinstance(ne, Wire) and not terminal_only:
                dist = Distance_From_Line(
                        ne.t1.Get_Location(), ne.t2.Get_Location(), (x, y))
                finder.append((dist * 0.5, ne))
        
        if len(finder) == 0:
            return None

        finder.sort()
        (dist, ne) = finder[ 0 ]
        return ne

    def Register(self, ne):
        assert isinstance(ne, Network_Element)
        self.new_elements.add(ne)
        if self.wxwin != None:
            ne.Refresh(self.wxwin)

    def Unregister(self, ne):
        assert isinstance(ne, Network_Element)
        if ne in self.new_elements:
            # Not actually added to the map yet...
            self.new_elements.remove(ne)
            assert not (ne in self.all_elements)
            return

        if self.wxwin != None:
            ne.Refresh(self.wxwin)

        self.map_contents_change = True
        self.all_elements.remove(ne)

        for (cx, cy) in ne.Get_Cells():
            assert self.cell_contents.has_key((cx, cy))

            #self.cell_contents[ (cx, cy) ].discard(ne)
            self.cell_contents[ (cx, cy) ].remove(ne)
            if len(self.cell_contents[ (cx, cy) ]) == 0:
                del self.cell_contents[ (cx, cy) ]

    def Update_Map(self, force_level):
        #print 'Update Map at level %d stage 0' % force_level

        while len(self.new_elements) != 0:
            ne = self.new_elements.pop()
            assert not (ne in self.all_elements)
            self.all_elements.add(ne)
            self.map_contents_change = True

            for (cx, cy) in ne.Get_Cells():
                if not self.cell_contents.has_key((cx, cy)):
                    self.cell_contents[ (cx, cy) ] = set()

                self.cell_contents[ (cx, cy) ].add(ne)

        if ((not self.map_contents_change)
        and (force_level <= 0)
        and not self.network_contents_change):
            return None

        #print 'Update Map at level %d stage 1' % force_level

        self.map_contents_change = False
        self.components = []
        self.terminals = []
        self.draw_before_components = []
        self.draw_after_components = []
        scrollbar_change = False

        try:
            for ne in self.all_elements:
                if isinstance(ne, Base_Component):
                    self.components.append(ne)

                    (x, y) = ne.Get_Location()
                    (w, h) = ne.Get_Size()
                    wh = w / 2
                    hh = h / 2

                    for (name, value) in [
                            ("mx1", - (x - 50 - wh)), ("my1", - (y - 50 - hh)),
                            ("x2", x + 50 + wh), ("y2", y + 50 + hh) ]:
                        if self.window_extent.get(name, -1000) < value:
                            self.window_extent[ name ] = value
                            scrollbar_change = True
                                
                elif isinstance(ne, Wire):
                    self.draw_before_components.append(ne)

                else:
                    self.draw_after_components.append(ne)
                    if isinstance(ne, Terminal):
                        self.terminals.append(ne)

            if scrollbar_change:
                if self.wxwin != None:
                    self.wxwin.Change_Scrollbar(**self.window_extent)

            if ((not self.network_contents_change)
            and (force_level <= 1)):
                self.components.sort(key = lambda n: n.depth)
                return None

            #print 'Update Map at level %d stage 2' % force_level
            self.network_contents_change = False

            for t in self.terminals:
                t.source = None

            for ne in self.components:
                ne.Reset_Depends()

            for ne in self.components:
                ne.Update_Depends()

            self.label_component = None
            roots = []
            for ne in self.components:
                if ((len(ne.provides) == 0)
                or isinstance(ne, Clocked_Component)):
                    roots.append(ne)

                if (isinstance(ne, Guide_Component)
                or isinstance(ne, Splash_Component)):
                    self.label_component = ne

            for root in roots:
                reached = set()

                def Calc_Depth(n, d):
                    if n in reached:
                        raise Combinatorial_Loop(n)

                    if (d + 1) <= n.depth:
                        # No change necessary
                        return

                    n.depth = d + 1
                    reached.add(n)
                    for n1 in n.provides:
                        if not isinstance(n1, Clocked_Component):
                            Calc_Depth(n1, d + 1)
                    reached.remove(n)
                        
                Calc_Depth(root, 0)

            for root in roots:
                if isinstance(root, D_Flip_Flop):
                    root.depth = len(self.components)
                
            self.components.sort(key = lambda n: n.depth)
            #print 'sorted:',
            #print '\n'.join([ str((n.depth, n)) for n in self.components ])
            #print 'no error'


        except Network_Error, e:
            self.network_contents_change = True
            #print 'error', e
            return e

        return None

    def __Simulate(self):
        rc = self.Update_Map(0)
        assert rc == None

        #print 'simulate'
        for ne in self.components:
            assert isinstance(ne, Base_Component)
            assert ne in self.all_elements
            if not isinstance(ne, Clocked_Component):
                ne.Set_State(ST_UNDEFINED)

        for ne in self.components:
            ne.Simulate()

        if self.wxwin != None:
            for ne in self.terminals:
                ne.Consider_Refresh(self.wxwin)

            #print ne, ne.depth, ne.Get_State()

    def Paint(self, dc):
        self.__Simulate()
        if self.selected_item != None:
            self.selected_item.Paint_Highlight(dc, self.wxp)
        if self.highlight != None:
            self.highlight.Paint_Highlight(dc, self.wxp)

        for nel in [self.draw_before_components, 
                    self.components, self.draw_after_components]:
            for ne in nel:
                ne.Paint(dc, self.wxp)

    def Popup_Done(self):
        rc = self.Update_Map(2)
        assert rc == None

        #self.popup_menu.Hide()
        self.network_contents_change = True
        self.popup_menu = None
        self.selected_item = self.highlight = None
        self.wxwin.Refresh()

    def Refresh_Highlight_and_SI(self):
        if self.wxwin == None:
            return
        if self.highlight != None:
            self.highlight.Refresh(self.wxwin)
        if self.selected_item != None:
            self.selected_item.Refresh(self.wxwin)

    def Mouse_Down(self, (x, y)):
        if self.mouse_down: 
            return False
        self.mouse_down = True

        if self.popup_menu != None:
            self.Popup_Done()

        self.Refresh_Highlight_and_SI()

        self.selected_item = self.highlight = self.Locate((x, y))
        self.scroll_point_at = None
        self.goal_major = None

        if isinstance(self.selected_item, Splash_Component):
            # selection denied
            self.selected_item = self.highlight = None

        elif isinstance(self.selected_item, Terminal):
            # Connect a new wire to this terminal
            self.creating_wire_dest = Temporary_Component(self, (x, y))
            self.creating_wire = Wire(self, 
                        self.selected_item, 
                        self.creating_wire_dest.Get_Terminal())
            self.Update_Map(2)

        elif isinstance(self.selected_item, Wire):
            # Pop up menu for Wire (Delete/Split/Cancel)
            (x1, y1) = Find_P4(self.selected_item.t1.Get_Location(), 
                    self.selected_item.t2.Get_Location(), (x, y))
            p4 = (int(x1), int(y1))
            self.popup_menu = self.wxwin.PopupMenu(
                    Wire_Popup_Menu(self.wxwin, self, p4), 
                    self.wxwin.Unoffset(p4))

        elif self.selected_item == None:
            # MGC-style scrolling
            self.scroll_point_at = (x, y)

        self.Refresh_Highlight_and_SI()
        
    def Mouse_Motion(self, (x, y)):
        self.Refresh_Highlight_and_SI()
        self.highlight = self.selected_item

        if self.selected_item == None: 
            # Highlight whatever the user is pointing at
            self.highlight = self.Locate((x, y))
            if (self.scroll_point_at != None) and self.mouse_down:
                (x2, y2) = self.scroll_point_at
                dx = x2 - x
                dy = y2 - y
                if ((dx != 0) or (dy != 0)):
                    self.wxwin.MGC_Scroll((dx, dy))


        elif isinstance(self.selected_item, Terminal):
            if self.mouse_down:
                # Drawing a new wire
                self.wxwin.Ensure_Visible((x, y))
                self.creating_wire_dest.Set_Location((x, y))
                self.highlight = self.Locate((x, y))
                if self.highlight == None:
                    # A new wire to a new location.
                    self.highlight = self.creating_wire_dest
                elif not isinstance(self.highlight, Terminal):
                    # I can't hit that!
                    self.highlight = None

                self.Update_Map(1)

        elif isinstance(self.selected_item, Base_Component):
            if self.mouse_down:
                # Dragging a component to a new location
                self.wxwin.Ensure_Visible((x, y))
                self.selected_item.Set_Location((x, y))
                self.Update_Map(1)

        self.Refresh_Highlight_and_SI()

    def Mouse_Up(self, (x, y)):
        if not self.mouse_down: 
            return False

        self.Refresh_Highlight_and_SI()
        error = None
        self.mouse_down = False
        self.highlight = None
        self.goal_major = None

        if self.selected_item == None: 
            pass
        
        elif isinstance(self.selected_item, Terminal):
            # Place new wire?
            final_target = self.Locate((x, y))
            new_wire = new_component = None
            wired_or_detect = False

            if ((final_target != None)
            and (final_target != self.selected_item)
            and (isinstance(final_target, Terminal))):
                
                # Target is ok... good
                # But will adding this wire make the network impossible
                # to simulate, necessitating a rollback?
                new_wire = Wire(self, self.selected_item, final_target) 
                wired_or_detect = (isinstance(final_target, Out_Terminal)
                        and isinstance(self.selected_item, Out_Terminal))

            elif (final_target == None):
                # Target is an empty space... good
                new_component = Split_Component(self, (x, y))
                new_wire = Wire(self, self.selected_item, 
                        new_component.Get_Terminal()) 

            assert self.creating_wire != None
            self.creating_wire.Delete()
            self.creating_wire_dest.Delete()
            self.creating_wire = None
            self.creating_wire_dest = None

            self.network_contents_change = True
            if wired_or_detect:
                error = Wired_OR(final_target)
            else:
                error = self.Update_Map(2)

            if error != None:
                assert new_wire != None
                print 'error: %s' % error
                new_wire.Delete()
                if new_component != None:
                    new_component.Delete()

                self.network_contents_change = True
                rc = self.Update_Map(2)
                assert rc == None

        elif isinstance(self.selected_item, Base_Component):
            self.selected_item.Set_Location((x, y))
            rc = self.Update_Map(1)
            assert rc == None

        self.selected_item = self.highlight = None

        if error != None:
            self.wxwin.Alert_Popup(error)

        self.Refresh_Highlight_and_SI()

    def Get_Output_Change(self):
        rc = False
        for (name, sd) in self.string_data_for_name.iteritems():
            if sd.state != sd.commit_state:
                if OVERPAINT_DEBUG:
                    print 'light %s change from %s to %s' % (name,
                            sd.commit_state, sd.state)
                sd.commit_state = sd.state
                rc = True

        return rc

    def Set_Strings(self, level_strings):
        self.string_data = []
        self.string_data_for_name = dict()
        self.win_flag = False

        for ls in level_strings:
            s = Dynamic_String_Data(ls.goal_sequence, ls.colour, ls.name) 
            assert not self.string_data_for_name.has_key(ls.name)
            self.string_data_for_name[ ls.name ] = s
            self.string_data.append(s)

        s = self.string_data[ 0 ]
        gsl = len(s.goal_sequence)
        assert gsl > 0

        for s in self.string_data:
            assert gsl == len(s.goal_sequence)

    def Get_String_Data(self):
        for s in self.string_data:
            yield s

    def Get_Number_of_Goals(self):
        s = self.string_data[ 0 ]
        gsl = len(s.goal_sequence)
        return gsl

    def Set_Light_State(self, name, state):
        if not self.string_data_for_name.has_key(name):
            # Unknown light
            return

        self.string_data_for_name[ name ].state = state

    def Get_Selected_Item(self):
        return self.selected_item

    def Is_Placing_Wire(self):
        return isinstance(self.selected_item, Terminal) and self.mouse_down

    def __Make_Stimulator_Sequence(self):
        # Make sure we have the list of components
        rc = self.Update_Map(0)
        assert rc == None

        # Find all stimulators
        self.stimulator_list = []
        period_set = set()
        for c in self.components:
            if isinstance(c, Stimulator):
                self.stimulator_list.append(c)
                assert len(c.sequence) > 0
                period_set.add(len(c.sequence))

        # Order must be sensible so that multiple calls to this
        # method don't break things
        self.stimulator_list.sort(key = lambda s: s.Get_Symbol())

        # Find hyperperiod (least common multiple of period set)
        period_set.discard(1)
        hyperperiod = Least_Common_Multiple(list(period_set))
        self.stimulator_sequence = []
        for i in xrange(hyperperiod):
            out = []
            for c in self.stimulator_list:
                assert len(c.sequence) <= hyperperiod
                assert (hyperperiod % len(c.sequence)) == 0
                out.append(c.sequence[ (i + self.clocks) % len(c.sequence) ])
            
            self.stimulator_sequence.append(out)

    def Get_Stimulator_Sequence(self):
        return self.stimulator_sequence

    def Get_Stimulator_Names(self):
        assert self.stimulator_list != None
        return [ c.Get_Symbol() for c in self.stimulator_list ]

    def Get_Stimulator_Major(self):
        assert self.stimulator_sequence != None
        return self.clocks % len(self.stimulator_sequence)

    def Update_Label(self, ack_counter):
        ne = self.label_component
        if ((self.wxwin == None) or (ne == None)):
            return { "location" : (0, 0) }

        return ne.Update(ack_counter)



class Dynamic_String_Data:
    def __init__(self, goal_sequence, colour, name):
        self.goal_sequence = goal_sequence
        self.colour = colour
        self.name = name
        self.state = ST_UNDEFINED
        self.commit_state = ST_UNDEFINED

class Dummy_Network(Network):
    # The dummy network synthesises the winning sequence.
    def __init__(self):
        Network.__init__(self)

    def Tick(self):
        if len(self.string_data) == 0:
            self.goal_major = None
            return
           
        gsl = self.Get_Number_of_Goals()

        if self.goal_major == None:
            self.goal_major = 0
        else:
            self.goal_major = (self.goal_major + 1) % gsl

        for s in self.string_data:
            self.Set_Light_State(s.name, 
                    s.goal_sequence[ self.goal_major ])


